/******************************************************************************
 *
 * Copyright (c) 2019, the jupyter-fs authors.
 *
 * This file is part of the jupyter-fs library, distributed under the terms of
 * the Apache License 2.0.  The full license can be found in the LICENSE file.
 *
 */
import { Button, Dialog, DialogTitle, DialogContent, DialogActions, ExpansionPanel, ExpansionPanelSummary, Typography, ExpansionPanelDetails, TextField, InputAdornment, IconButton, Tooltip, } from "@material-ui/core";
import * as React from "react";
import { visibilityIcon, visibilityOffIcon } from "./icons";
class Template {
    constructor(template) {
        this.template = template;
        this.pattern = /(?:)/;
    }
    substitute(dict) {
        return this.template.replace(this.pattern, (_, p1) => dict[p1]);
    }
    tokens() {
        const toks = [];
        let match;
        // eslint-disable-next-line no-cond-assign
        while ((match = this.pattern.exec(this.template))) {
            toks.push(match[1]);
        }
        return toks;
    }
}
export class DoubleBraceTemplate extends Template {
    constructor() {
        super(...arguments);
        this.pattern = /{{(\S+?)}}/g;
    }
}
function tokensFromUrl(url) {
    return new DoubleBraceTemplate(url).tokens();
}
function _askRequired(spec) {
    var _a;
    return spec.auth === "ask" && !spec.init && !!((_a = spec.missingTokens) === null || _a === void 0 ? void 0 : _a.length);
}
export function askRequired(specs) {
    for (const spec of specs) {
        if (_askRequired(spec)) {
            return true;
        }
    }
    return false;
}
export class AskDialog extends React.Component {
    constructor(props, options) {
        super(props);
        this.state = AskDialog.initialState({ open: true, options });
    }
    render() {
        return (React.createElement("div", null,
            React.createElement(Dialog, { className: "jfs-ask-dialog", open: this.state.open, onClose: this._onClose.bind(this) },
                React.createElement(DialogTitle, { className: "jfs-ask jfs-ask-dialog-title" }, "Please enter token values for filesystem resources"),
                React.createElement(DialogContent, { className: "jfs-ask jfs-ask-dialog-content" }, this._form()),
                React.createElement(DialogActions, { className: "jfs-ask jfs-ask-dialog-actions" },
                    React.createElement(Button, { className: "jfs-ask jfs-ask-dialog-actions-button-cancel", onClick: this._onClose.bind(this), color: "primary" }, "Cancel"),
                    React.createElement(Button, { className: "jfs-ask jfs-ask-dialog-actions-button-submit", onClick: this._onSubmit.bind(this), color: "primary" }, "Submit")))));
    }
    _form() {
        return (React.createElement("form", { className: "jfs-ask jfs-ask-form", onSubmit: this._onSubmit.bind(this), noValidate: true, autoComplete: "off" }, this._formInner()));
    }
    _formInner() {
        return this.props.resources.map(resource => {
            // ask for credentials if needed, or state why not
            const decodedUrl = decodeURIComponent(resource.url);
            const askReq = _askRequired(resource);
            const inputs = askReq ? this._inputs(resource.url) : [];
            const tokens = tokensFromUrl(resource.url);
            let reason = "";
            if (resource.init && this.props.options.cache) {
                reason = "already initialized";
            }
            else if (!tokens.length) {
                reason = "no template parameters";
            }
            const summary = `${resource.name}:${reason && ` ${reason}`}`;
            return [
                React.createElement(ExpansionPanel, { className: "jfs-ask jfs-ask-panel", disabled: !!reason, expanded: !reason, key: `${resource.name}_panel` },
                    React.createElement(ExpansionPanelSummary, { className: "jfs-ask jfs-ask-panel-summary" },
                        React.createElement(Typography, null, summary),
                        !reason &&
                            React.createElement(Tooltip, { title: decodedUrl },
                                React.createElement(Typography, { noWrap: true }, decodedUrl))),
                    React.createElement(ExpansionPanelDetails, { className: "jfs-ask jfs-ask-panel-details" }, inputs)),
            ];
        });
    }
    _inputs(url) {
        return tokensFromUrl(url).map(token => {
            var _a, _b, _c;
            return (React.createElement(TextField, { className: "jfs-ask jfs-ask-input", autoFocus: true, fullWidth: true, key: `${url}_${token}`, label: token, margin: "dense", name: token, onChange: this._onChange(url).bind(this), type: ((_a = this.state.visibility[url]) === null || _a === void 0 ? void 0 : _a[token]) ? "text" : "password", value: ((_b = this.state.values[url]) === null || _b === void 0 ? void 0 : _b[token]) || "", InputProps: {
                    endAdornment: (React.createElement(InputAdornment, { position: "end" },
                        React.createElement(IconButton, { onClick: this._onClickVisiblity(url, token).bind(this), onMouseDown: this._onMouseDownVisibility.bind(this), edge: "end" }, ((_c = this.state.visibility[url]) === null || _c === void 0 ? void 0 : _c[token]) ? (React.createElement(visibilityIcon.react, null)) : (React.createElement(visibilityOffIcon.react, null))))),
                } }));
        });
    }
    _onChange(url) {
        return function (event) {
            const target = event.target;
            this._setValue(url, target.name, target.value);
        };
    }
    _onClickVisiblity(url, key) {
        return function () {
            this._toggleVisibility(url, key);
        };
    }
    _onMouseDownVisibility(event) {
        event.preventDefault();
    }
    _onOpen() {
        this.setState({ open: true });
    }
    _onClose() {
        // close the dialog and blank the form
        this.setState(AskDialog.initialState({ options: this.state.options }));
        this.props.handleClose();
    }
    async _onSubmit(event) {
        event.preventDefault();
        this.props.handleSubmit(this.state.values);
        this._onClose();
    }
    _setValue(url, key, value) {
        const urlValues = { ...this.state.values[url], [key]: value };
        this.setState({
            values: { ...this.state.values, [url]: urlValues },
        });
    }
    _toggleVisibility(url, key) {
        var _a;
        const urlVis = {
            ...this.state.visibility[url],
            [key]: !((_a = this.state.visibility[url]) === null || _a === void 0 ? void 0 : _a[key]),
        };
        this.setState({
            visibility: { ...this.state.visibility, [url]: urlVis },
        });
    }
}
AskDialog.displayName = "AskDialog";
/**
 * A namespace for AskDialog statics.
 */
(function (AskDialog) {
    /**
     * The initial state for a new AskDialog
     */
    AskDialog.initialState = ({ open = false, options, }) => ({
        open,
        options: { ...options },
        values: {},
        visibility: {},
    });
})(AskDialog || (AskDialog = {}));
//# sourceMappingURL=auth.js.map