/******************************************************************************
 *
 * Copyright (c) 2019, the jupyter-fs authors.
 *
 * This file is part of the jupyter-fs library, distributed under the terms of
 * the Apache License 2.0.  The full license can be found in the LICENSE file.
 *
 */
/* eslint-disable @typescript-eslint/no-non-null-assertion */
import { Dialog, showDialog } from "@jupyterlab/apputils";
import { closeIcon, copyIcon, cutIcon, downloadIcon, editIcon, filterListIcon, pasteIcon, refreshIcon, fileIcon, newFolderIcon, } from "@jupyterlab/ui-components";
import { map } from "@lumino/algorithm";
import { DisposableSet } from "@lumino/disposable";
import { Menu } from "@lumino/widgets";
import { Path } from "@tree-finder/base";
import { TreeFinderSidebar } from "./treefinder";
import { getContentParent, getRefreshTargets, openDirRecursive, revealAndSelectPath, splitPathstrDrive } from "./contents_utils";
import { showErrorMessage } from "@jupyterlab/apputils";
import { getAllSnippets, instantiateSnippet } from "./snippets";
// define the command ids as a constant tuple
export const commandNames = [
    "copy",
    "cut",
    "delete",
    "open",
    "openWith",
    "paste",
    "refresh",
    "rename",
    "download",
    "create_folder",
    "create_file",
    // "navigate",
    "copyFullPath",
    "copyRelativePath",
    "restore",
    "toggleColumnPath",
    "toggleColumn",
];
export const commandIDs = Object.fromEntries(commandNames.map(name => [name, `treefinder:${name}`]));
const COLUMN_NAMES = [
    "size",
    "last_modified",
    "writable",
    "mimetype",
];
export function idFromResource(resource) {
    return [resource.name.split(" ").join(""), resource.drive].join("_");
}
const currentWidgetSelectionIsWritable = (tracker) => {
    var _a;
    if (!tracker.currentWidget) {
        return false;
    }
    const selection = (_a = tracker.currentWidget.treefinder.model) === null || _a === void 0 ? void 0 : _a.selection;
    if (selection) {
        return selection.every((x) => x.row.writable);
    }
    return false;
};
function toggleColumnCommandId(column) {
    return `${commandIDs.toggleColumn}-${column}`;
}
function _getRelativePaths(selectedFiles) {
    const allPaths = [];
    for (const file of selectedFiles) {
        const relativePath = file.getPathAtDepth(1).join("/");
        allPaths.push(relativePath);
    }
    return allPaths;
}
async function _digestString(value) {
    const encoded = new TextEncoder().encode(value); // encode as (utf-8) Uint8Array
    const buffer = await crypto.subtle.digest("SHA-256", encoded); // hash the message
    const hash = Array.from(new Uint8Array(buffer)); // convert buffer to byte array
    return hash
        .map(b => b.toString(16).padStart(2, "0"))
        .join(""); // convert bytes to hex string
}
async function _commandKeyForSnippet(snippet) {
    return `jupyterfs:snippet-${snippet.label}-${await _digestString(snippet.label + snippet.caption + snippet.pattern.source + snippet.template)}`;
}
function _openWithKeyForFactory(factory) {
    return `jupyterfs:openwith-${factory}`;
}
function _normalizedUrlForSnippet(content, baseUrl) {
    const path = splitPathstrDrive(content.pathstr)[1];
    return `${baseUrl}/${path}${content.hasChildren ? "/" : ""}`;
}
/**
 * Create commands that will have the same IDs indepent of settings/resources
 *
 * These commands do not need to be recreated on settings/resource updates
 */
export function createStaticCommands(app, tracker, clipboard) {
    return [
        app.commands.addCommand(commandIDs.copy, {
            execute: args => clipboard.model.copySelection(tracker.currentWidget.treefinder.model),
            icon: copyIcon,
            label: "Copy",
            isVisible: () => {
                var _a;
                const widget = tracker.currentWidget;
                if (!widget) {
                    return false;
                }
                // Copy of folders are unsupported
                if ((_a = widget.treefinder.model) === null || _a === void 0 ? void 0 : _a.selection.some(v => v.row.kind === "dir")) {
                    return false;
                }
                return true;
            },
            isEnabled: () => !!tracker.currentWidget,
        }),
        app.commands.addCommand(commandIDs.cut, {
            execute: args => clipboard.model.cutSelection(tracker.currentWidget.treefinder.model),
            icon: cutIcon,
            label: "Cut",
            isEnabled: () => currentWidgetSelectionIsWritable(tracker),
        }),
        app.commands.addCommand(commandIDs.delete, {
            execute: async (args) => {
                const treefinder = tracker.currentWidget.treefinder;
                const model = treefinder.model;
                const message = model.selection.length === 1
                    ? `Are you sure you want to permanently delete: ${model.selection[0].name}?`
                    : `Are you sure you want to permanently delete the ${model.selection.length} selected items?`;
                const result = await showDialog({
                    title: "Delete",
                    body: message,
                    buttons: [
                        Dialog.cancelButton({ label: "Cancel" }),
                        Dialog.warnButton({ label: "Delete" }),
                    ],
                    // By default focus on "Cancel" to protect from accidental deletion
                    defaultButton: 0,
                });
                if (!treefinder.isDisposed && result.button.accept) {
                    clipboard.model.deleteSelection(model);
                }
            },
            icon: closeIcon.bindprops({ stylesheet: "menuItem" }),
            label: "Delete",
            isEnabled: () => currentWidgetSelectionIsWritable(tracker),
        }),
        app.commands.addCommand(commandIDs.open, {
            execute: args => { var _a; return tracker.currentWidget.treefinder.model.openSub.next(((_a = tracker.currentWidget.treefinder.selection) === null || _a === void 0 ? void 0 : _a.map(c => c.row)) || []); },
            label: "Open",
            isEnabled: () => !!tracker.currentWidget,
        }),
        app.commands.addCommand(commandIDs.paste, {
            execute: args => clipboard.model.pasteSelection(tracker.currentWidget.treefinder.model),
            icon: pasteIcon,
            label: "Paste",
            isEnabled: () => !!tracker.currentWidget,
        }),
        app.commands.addCommand(commandIDs.rename, {
            execute: args => {
                const widget = tracker.currentWidget;
                const oldContent = widget.treefinder.selection[0];
                void TreeFinderSidebar.doRename(widget, oldContent).then(newContent => {
                    var _a;
                    (_a = widget.treefinder.model) === null || _a === void 0 ? void 0 : _a.renamerSub.next({ name: newContent.name, target: oldContent });
                    // TODO: Model state of TreeFinderWidget should be updated by renamerSub process.
                    oldContent.row = newContent;
                });
            },
            icon: editIcon,
            label: "Rename",
            isEnabled: () => currentWidgetSelectionIsWritable(tracker),
        }),
        app.commands.addCommand(commandIDs.download, {
            execute: async (args) => {
                const widget = tracker.currentWidget;
                const selection = widget.treefinder.selection;
                await Promise.allSettled(selection.map(s => widget.download(s.pathstr, s.hasChildren)));
            },
            icon: downloadIcon,
            label: "Download",
            isEnabled: () => {
                var _a, _b, _c;
                return !!((_c = (_b = (_a = tracker.currentWidget) === null || _a === void 0 ? void 0 : _a.treefinder.model) === null || _b === void 0 ? void 0 : _b.selection) === null || _c === void 0 ? void 0 : _c.some(s => !s.hasChildren));
            },
        }),
        app.commands.addCommand(commandIDs.create_folder, {
            execute: async (args) => {
                var _a;
                const widget = tracker.currentWidget;
                const model = widget.treefinder.model;
                let target = (_a = model.selectedLast) !== null && _a !== void 0 ? _a : model.root;
                if (!target.hasChildren) {
                    target = await getContentParent(target, model.root);
                }
                const path = Path.fromarray(target.row.path);
                let row;
                try {
                    row = await widget.treefinder.contentsProxy.newUntitled({
                        type: "directory",
                        path,
                    });
                }
                catch (e) {
                    void showErrorMessage("Could not create folder", e);
                    return;
                }
                target.invalidate();
                const content = await revealAndSelectPath(model, row.path);
                // Is this really needed?
                model.refreshSub.next(getRefreshTargets([target.row], model.root) || []);
                // Scroll into view if not visible
                await TreeFinderSidebar.scrollIntoView(widget.treefinder, content.pathstr);
                const newContent = await TreeFinderSidebar.doRename(widget, content);
                model.renamerSub.next({ name: newContent.name, target: content });
                // TODO: Model state of TreeFinderWidget should be updated by renamerSub process.
                content.row = newContent;
            },
            icon: newFolderIcon,
            label: "New Folder",
            isEnabled: () => !!tracker.currentWidget,
        }),
        app.commands.addCommand(commandIDs.create_file, {
            execute: async (args) => {
                var _a;
                const widget = tracker.currentWidget;
                const model = widget.treefinder.model;
                let target = (_a = model.selectedLast) !== null && _a !== void 0 ? _a : model.root;
                if (!target.hasChildren) {
                    target = await getContentParent(target, model.root);
                }
                const path = Path.fromarray(target.row.path);
                let row;
                try {
                    row = await widget.treefinder.contentsProxy.newUntitled({
                        type: "file",
                        path,
                    });
                }
                catch (e) {
                    void showErrorMessage("Could not create file", e);
                    return;
                }
                target.invalidate();
                const content = await revealAndSelectPath(model, row.path);
                // Is this really needed?
                model.refreshSub.next(getRefreshTargets([target.row], model.root) || []);
                // Scroll into view if not visible
                await TreeFinderSidebar.scrollIntoView(widget.treefinder, content.pathstr);
                const newContent = await TreeFinderSidebar.doRename(widget, content);
                model.renamerSub.next({ name: newContent.name, target: content });
                // TODO: Model state of TreeFinderWidget should be updated by renamerSub process.
                content.row = newContent;
            },
            icon: fileIcon,
            label: "New File",
            isEnabled: () => !!tracker.currentWidget,
        }),
        app.commands.addCommand(commandIDs.refresh, {
            execute: args => {
                if (args["selection"]) {
                    clipboard.refreshSelection(tracker.currentWidget.treefinder.model);
                }
                else {
                    clipboard.refresh(tracker.currentWidget.treefinder.model);
                }
            },
            icon: refreshIcon,
            label: args => args["selection"] ? "Refresh Selection" : "Refresh",
            isEnabled: () => !!tracker.currentWidget,
        }),
        app.commands.addCommand(commandIDs.copyFullPath, {
            execute: async (args) => {
                const widget = tracker.currentWidget;
                const trimEnd = (path) => path.trimEnd().replace(/\/+$/, "");
                const fullPaths = _getRelativePaths(widget.treefinder.selection).map(relativePath => { var _a; return [trimEnd((_a = widget.url) !== null && _a !== void 0 ? _a : ""), relativePath].join("/"); });
                await navigator.clipboard.writeText(fullPaths.join("\n"));
            },
            label: "Copy Full Path",
            isEnabled: () => !!tracker.currentWidget,
        }),
        app.commands.addCommand(commandIDs.copyRelativePath, {
            execute: async (args) => {
                const widget = tracker.currentWidget;
                const relativePaths = _getRelativePaths(widget.treefinder.selection);
                await navigator.clipboard.writeText(relativePaths.join("\n"));
            },
            label: "Copy Relative Path",
            isEnabled: () => !!tracker.currentWidget,
        }),
        app.commands.addCommand(commandIDs.toggleColumnPath, {
            execute: args => { },
            label: "path",
            isEnabled: () => false,
            isToggled: () => true,
        }),
        app.commands.addCommand(commandIDs.restore, {
            execute: async (args) => {
                const rootPath = args.rootPath;
                const dirsToOpen = rootPath.split("/");
                const sidebar = tracker.findByDrive(args.id);
                if (!sidebar) {
                    throw new Error(`Could not restore JupyterFS browser: ${args.id}`);
                }
                const treefinderwidget = sidebar.treefinder;
                const model = treefinderwidget.model;
                // If preferredDir not specified, proceed with the restore
                if (!sidebar.preferredDir) {
                    await openDirRecursive(model, dirsToOpen);
                    await tracker.save(sidebar);
                }
            },
        }),
    ].reduce((set, d) => {
        set.add(d);
        return set;
    }, new DisposableSet());
}
/**
 * Create commands whose count/IDs depend on settings/resources
 */
export async function createDynamicCommands(app, tracker, clipboard, resources, settings) {
    var _a;
    const columnCommands = [];
    const toggleState = {};
    const colsToDisplay = (_a = settings === null || settings === void 0 ? void 0 : settings.composite.display_columns) !== null && _a !== void 0 ? _a : ["size"];
    const columnsMenu = new Menu({ commands: app.commands });
    columnsMenu.title.label = "Show/Hide Columns";
    columnsMenu.title.icon = filterListIcon;
    columnsMenu.addItem({ command: commandIDs.toggleColumnPath });
    for (const column of COLUMN_NAMES) {
        columnsMenu.addItem({ command: toggleColumnCommandId(column) });
        toggleState[column] = colsToDisplay.includes(column);
        columnCommands.push(app.commands.addCommand(toggleColumnCommandId(column), {
            execute: async (args) => {
                toggleState[column] = !toggleState[column];
                await (settings === null || settings === void 0 ? void 0 : settings.set("display_columns", COLUMN_NAMES.filter(k => toggleState[k])));
            },
            label: column,
            isToggleable: true,
            isToggled: () => toggleState[column],
        }));
    }
    const snippetsMenu = new Menu({ commands: app.commands });
    snippetsMenu.title.label = "Snippets";
    const snippets = await getAllSnippets(settings);
    const snippetCommands = [];
    const snippetIds = new Set();
    for (const snippet of snippets) {
        const key = await _commandKeyForSnippet(snippet);
        if (snippetIds.has(key)) {
            console.warn("Discarding duplicate snippet", snippet);
            continue;
        }
        snippetIds.add(key);
        snippetsMenu.addItem({ command: key });
        snippetCommands.push(app.commands.addCommand(key, {
            execute: async (args) => {
                const sidebar = tracker.currentWidget;
                const content = sidebar.treefinder.selection[0];
                const instantiated = instantiateSnippet(snippet.template, sidebar.url, sidebar.type, content.pathstr);
                await navigator.clipboard.writeText(instantiated);
            },
            label: snippet.label,
            caption: snippet.caption,
            isVisible: () => {
                const sidebar = tracker.currentWidget;
                const selection = sidebar === null || sidebar === void 0 ? void 0 : sidebar.treefinder.selection;
                // discard if not for backend type
                if (snippet.type !== "" && snippet.type !== (sidebar === null || sidebar === void 0 ? void 0 : sidebar.type)) {
                    return false;
                }
                // include if matches pattern
                if (selection === null || selection === void 0 ? void 0 : selection.length) {
                    return snippet.pattern.test(_normalizedUrlForSnippet(selection[0], sidebar.url));
                }
                return false;
            },
        }));
    }
    const openWithCommands = [];
    const openWithMenu = new Menu({ commands: app.commands });
    openWithMenu.title.label = "Open With";
    openWithMenu.id = "treefinder:open-with";
    const { docRegistry } = app;
    let items = [];
    function updateOpenWithMenu(contextMenu) {
        var _a, _b;
        const openWith = (_b = (_a = contextMenu.menu.items.find(item => {
            var _a;
            return item.type === "submenu" &&
                ((_a = item.submenu) === null || _a === void 0 ? void 0 : _a.id) === "treefinder:open-with";
        })) === null || _a === void 0 ? void 0 : _a.submenu) !== null && _b !== void 0 ? _b : null;
        if (!openWith) {
            return; // Bail early if the open with menu is not displayed
        }
        // clear the current menu items
        // items.forEach(item => item.dispose());
        items.length = 0;
        // Ensure that the menu is empty
        openWith.clearItems();
        // clear the commands
        openWithCommands.forEach(item => item.dispose());
        openWithCommands.length = 0;
        // get the widget factories that could be used to open all of the items
        // in the current filebrowser selection
        const widget = tracker.currentWidget;
        const treefinder = widget.treefinder;
        const model = treefinder.model;
        const factories = tracker.currentWidget
            ? intersection(map(model.selection, i => getFactories(docRegistry, widget, i)))
            : new Set();
        // make new menu items from the widget factories
        items = [...factories].map(factory => {
            const key = _openWithKeyForFactory(factory.label || factory.name);
            const label = factory.label || factory.name;
            openWithCommands.push(app.commands.addCommand(key, {
                execute: args => Promise.all(Array.from(map(model.selection, item => app.commands.execute("docmanager:open", { path: Path.fromarray(item.row.path), ...args })))),
                label,
                isVisible: () => true,
            }));
            return openWith.addItem({
                args: { factory: factory.name, label: factory.label || factory.name },
                command: key,
            });
        });
    }
    app.contextMenu.opened.connect(updateOpenWithMenu);
    const selector = ".jp-tree-finder-sidebar";
    let contextMenuRank = 1;
    return [
        ...columnCommands,
        ...snippetCommands,
        // context menu items
        app.contextMenu.addItem({
            command: commandIDs.open,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            type: "submenu",
            submenu: openWithMenu,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.copy,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.cut,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.paste,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.delete,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.rename,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.download,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            type: "separator",
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.copyFullPath,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.copyRelativePath,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            type: "submenu",
            submenu: snippetsMenu,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            type: "separator",
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.create_file,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            command: commandIDs.create_folder,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            type: "separator",
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            type: "submenu",
            submenu: columnsMenu,
            selector,
            rank: contextMenuRank++,
        }),
        app.contextMenu.addItem({
            args: { selection: true },
            command: commandIDs.refresh,
            selector,
            rank: contextMenuRank++,
        }),
    ].reduce((set, d) => {
        set.add(d);
        return set;
    }, new DisposableSet());
}
export function getFactories(docRegistry, widget, item) {
    const path = [widget.url.trimEnd().replace(/\/+$/, ""), item.getPathAtDepth(1).join("/")].join("/");
    const factories = docRegistry.preferredWidgetFactories(path);
    const notebookFactory = docRegistry.getWidgetFactory("notebook");
    if (notebookFactory &&
        item.row.kind === "notebook" &&
        factories.indexOf(notebookFactory) === -1) {
        factories.unshift(notebookFactory);
    }
    return factories;
}
function intersection(iterables) {
    let accumulator;
    for (const current of iterables) {
        // Initialize accumulator.
        if (accumulator === undefined) {
            accumulator = new Set(current);
            continue;
        }
        // Return early if empty.
        if (accumulator.size === 0) {
            return accumulator;
        }
        // Keep the intersection of accumulator and current.
        const intersection_set = new Set();
        for (const value of current) {
            if (accumulator.has(value)) {
                intersection_set.add(value);
            }
        }
        accumulator = intersection_set;
    }
    return accumulator !== null && accumulator !== void 0 ? accumulator : new Set();
}
//# sourceMappingURL=commands.js.map