import { PromiseDelegate } from "@lumino/coreutils";
import { showErrorMessage } from "@jupyterlab/apputils";
import { Path } from "@tree-finder/base";
/**
 * Wrapper for a drive onto the contents manager.
 */
export class ContentsProxy {
    constructor(contentsManager, drive, onGetChildren) {
        this.contentsManager = contentsManager;
        this.drive = drive;
        this.onGetChildren = onGetChildren;
    }
    async get(path, options) {
        path = ContentsProxy.toFullPath(path, this.drive);
        return ContentsProxy.toJupyterContentRow(await this.contentsManager.get(path, options), this.contentsManager, this.drive, this.onGetChildren);
    }
    async save(path, options) {
        path = ContentsProxy.toFullPath(path, this.drive);
        return ContentsProxy.toJupyterContentRow(await this.contentsManager.save(path, options), this.contentsManager, this.drive, this.onGetChildren);
    }
    async rename(path, newPath) {
        path = ContentsProxy.toFullPath(path, this.drive);
        newPath = ContentsProxy.toFullPath(newPath, this.drive);
        return ContentsProxy.toJupyterContentRow(await this.contentsManager.rename(path, newPath), this.contentsManager, this.drive, this.onGetChildren);
    }
    async newUntitled(options) {
        options.path = options.path && ContentsProxy.toFullPath(options.path, this.drive);
        return ContentsProxy.toJupyterContentRow(await this.contentsManager.newUntitled(options), this.contentsManager, this.drive, this.onGetChildren);
    }
    async downloadUrl(path) {
        path = ContentsProxy.toFullPath(path, this.drive);
        return await this.contentsManager.getDownloadUrl(path);
    }
}
(function (ContentsProxy) {
    function toFullPath(path, drive) {
        if (!drive || path.startsWith(`${drive}:`)) {
            if (path.startsWith(`${drive}:/`)) {
                return path.replace(`${drive}:/`, `${drive}:`);
            }
            return path;
        }
        else if (path.startsWith(`${drive}/`)) {
            return [drive, path.slice(drive.length + 1)].join(":");
        }
        else {
            return [drive, path].join(":");
        }
    }
    ContentsProxy.toFullPath = toFullPath;
    function toLocalPath(path) {
        const [first, ...rest] = path.split("/");
        return [first.split(":").pop(), ...rest].join("/");
    }
    ContentsProxy.toLocalPath = toLocalPath;
    function toJupyterContentRow(row, contentsManager, drive, onGetChildren) {
        const { path, type, ...rest } = row;
        const pathWithDrive = toFullPath(path, drive).replace(/\/$/, "");
        const kind = type === "directory" ? "dir" : type;
        return {
            path: Path.toarray(pathWithDrive),
            kind,
            ...rest,
            ...(kind === "dir" ? {
                getChildren: async () => {
                    let contents;
                    const done = new PromiseDelegate();
                    if (onGetChildren) {
                        const pathstr = Path.toarray(pathWithDrive).join("/"); // maybe clean up the different path formats we have...
                        onGetChildren(pathstr, done.promise);
                    }
                    try {
                        contents = await contentsManager.get(pathWithDrive, { content: true });
                        done.resolve();
                    }
                    catch (error) {
                        void showErrorMessage("Failed to get directory contents", error);
                        done.reject("Failed to get directory contents");
                        return [];
                    }
                    return contents.content.map(c => toJupyterContentRow(c, contentsManager, drive, onGetChildren));
                },
            } : {}),
        };
    }
    ContentsProxy.toJupyterContentRow = toJupyterContentRow;
})(ContentsProxy || (ContentsProxy = {}));
//# sourceMappingURL=contents_proxy.js.map