/******************************************************************************
 *
 * Copyright (c) 2019, the jupyter-fs authors.
 *
 * This file is part of the jupyter-fs library, distributed under the terms of
 * the Apache License 2.0.  The full license can be found in the LICENSE file.
 *
 */
import { Widget } from "@lumino/widgets";
import type { Message } from "@lumino/messaging";
import { MimeData } from "@lumino/coreutils";
import { Drag } from "@lumino/dragdrop";
/**
 * The mime type for a table header drag object.
 */
export declare const TABLE_HEADER_MIME = "application/x-jupyterfs-thead";
/**
 * Determine whether node is equal to or a descendant of our widget, and that it does
 * not belong to a nested drag widget.
 */
export declare function belongsToUs(node: HTMLElement, parentClass: string, parentNode: HTMLElement): boolean;
/**
 * Find the direct child node of `parent`, which has `node` as a descendant.
 * Alternatively, parent can be a collection of children.
 *
 * Returns null if not found.
 */
export declare function findChild(parent: HTMLElement | HTMLElement[], node: HTMLElement): HTMLElement | null;
/**
 * A widget class which allows the user to drop mime data onto it.
 *
 * To complete the class, the following functions need to be implemented:
 *  - processDrop: Process pre-screened drop events
 *
 * The functionallity of the class can be extended by overriding the following
 * functions:
 *  - findDropTarget(): Override if anything other than the direct children
 *    of the widget's node are to be the drop targets.
 *
 * For maximum control, `evtDrop` can be overriden.
 */
export declare abstract class DropWidget extends Widget {
    /**
     * Construct a drop widget.
     */
    constructor(options?: DropWidget.IOptions);
    /**
     * Whether the widget should accept drops from an external source,
     * or only accept drops from itself.
     * Defaults to false, which will disallow all drops unless widget
     * is also a drag widget.
     */
    acceptDropsFromExternalSource: boolean;
    /**
     * Which mimetypes that the widget accepts for drops
     */
    acceptedDropMimeTypes: string[];
    /**
     * Handle the DOM events for the widget.
     *
     * @param event - The DOM event sent to the widget.
     *
     * #### Notes
     * This method implements the DOM `EventListener` interface and is
     * called in response to events on the drop widget's node. It should
     * not be called directly by user code.
     */
    handleEvent(event: Event): void;
    protected validateSource(event: Drag.Event): boolean;
    /**
     * Processes a drop event.
     *
     * This function is called after checking:
     *  - That the `dropTarget` is a valid drop target
     *  - The value of `event.source` if `acceptDropsFromExternalSource` is false
     */
    protected abstract processDrop(dropTarget: HTMLElement, event: Drag.Event): void;
    /**
     * Find a drop target from a given drag event target.
     *
     * Returns null if no valid drop target was found.
     *
     * The default implementation returns the direct child that is the parent of
     * `node`, or `node` if it is itself a direct child. It also checks that the
     * needed mime type is included
     */
    protected findDropTarget(input: HTMLElement, mimeData: MimeData): HTMLElement | null;
    /**
     * Handle the `'lm-drop'` event for the widget.
     *
     * Responsible for pre-processing event before calling `processDrop`.
     *
     * Should normally only be overriden if you cannot achive your goal by
     * other overrides.
     */
    protected evtDrop(event: Drag.Event): void;
    /**
     * Handle `after_attach` messages for the widget.
     */
    protected onAfterAttach(msg: Message): void;
    /**
     * Handle `before_detach` messages for the widget.
     */
    protected onBeforeDetach(msg: Message): void;
    /**
     * Handle the `'lm-dragenter'` event for the widget.
     */
    private _evtDragEnter;
    /**
     * Handle the `'lm-dragleave'` event for the widget.
     */
    private _evtDragLeave;
    /**
     * Handle the `'lm-dragover'` event for the widget.
     */
    private _evtDragOver;
    /**
     * Checks if the target is a header, and it is not 'path' or itself
     */
    private _isValidTargetHeader;
    /**
     * Clear existing drop target from out children.
     *
     * #### Notes
     * This function assumes there are only one active drop target
     */
    private _clearDropTarget;
}
/**
 * An internal base class for implementing drag operations on top
 * of drop class.
 */
export declare abstract class DragDropWidgetBase extends DropWidget {
    /**
     * Construct a drag and drop base widget.
     */
    constructor(options?: DragDropWidget.IOptions);
    /**
     * Dispose of the resources held by the directory listing.
     */
    dispose(): void;
    /**
     * Handle the DOM events for the widget.
     *
     * @param event - The DOM event sent to the widget.
     *
     * #### Notes
     * This method implements the DOM `EventListener` interface and is
     * called in response to events on the drag widget's node. It should
     * not be called directly by user code.
     */
    handleEvent(event: Event): void;
    /**
     * Adds mime data representing the drag data to the drag event's MimeData bundle.
     */
    protected abstract addMimeData(handle: HTMLElement, mimeData: MimeData): void;
    /**
     * Finds the drag target (the node to move) from a drag handle.
     *
     * Returns null if no valid drag target was found.
     *
     * The default implementation returns the handle directly.
     */
    protected findDragTarget(handle: HTMLElement): HTMLElement | null;
    /**
     * Returns the drag image to use when dragging using the given handle.
     *
     * The default implementation returns a clone of the drag target.
     */
    protected getDragImage(handle: HTMLElement): HTMLElement | null;
    /**
     * Called when a drag has completed with this widget as a source
     */
    protected onDragComplete(action: Drag.DropAction): void;
    /**
     * Handle `after_attach` messages for the widget.
     */
    protected onAfterAttach(msg: Message): void;
    /**
     * Handle `before_detach` messages for the widget.
     */
    protected onBeforeDetach(msg: Message): void;
    /**
     * Start a drag event.
     *
     * Called when dragging and DRAG_THRESHOLD is met.
     *
     * Should normally only be overriden if you cannot achieve your goal by
     * other overrides.
     */
    protected startDrag(handle: HTMLElement, clientX: number, clientY: number): void;
    /**
     * Drag data stored in _startDrag
     */
    protected drag: Drag | null;
    protected dragHandleClass: string;
    /**
     * Check if node, or any of nodes ancestors are a drag handle
     *
     * If it is a drag handle, it returns the handle, if not returns null.
     */
    private _findDragHandle;
    /**
     * Handle the `'mousedown'` event for the widget.
     */
    private _evtDragMousedown;
    /**
     * Handle the `'mouseup'` event for the widget.
     */
    private _evtDragMouseup;
    /**
     * Handle the `'mousemove'` event for the widget.
     */
    private _evtDragMousemove;
    protected defaultSupportedActions: Drag.SupportedActions;
    protected defaultProposedAction: Drag.DropAction;
    /**
     * Data stored on mouse down to determine if drag treshold has
     * been overcome, and to initialize drag once it has.
     */
    private _clickData;
}
/**
 * A widget which allows the user to initiate drag operations.
 *
 * Any descendant element with the drag handle class `'jfs-mod-dragHandle'`
 * will serve as a handle that can be used for dragging. If DragWidgets are
 * nested, handles will only belong to the closest parent DragWidget. For
 * convenience, the functions `makeHandle`, `unmakeHandle` and
 * `createDefaultHandle` can be used to indicate which elements should be
 * made handles. `createDefaultHandle` will create a new element as a handle
 * with a default styling class applied. Optionally, `childrenAreDragHandles`
 * can be set to indicate that all direct children are themselve drag handles.
 *
 * To complete the class, the following functions need to be implemented:
 * - addMimeData: Adds mime data to new drag events
 *
 * The functionallity of the class can be extended by overriding the following
 * functions:
 *  - findDragTarget(): Override if anything other than the direct children
 *    of the widget's node are to be drag targets.
 *  - getDragImage: Override to change the drag image (the default is a
 *    copy of the drag target).
 *  - onDragComplete(): Callback on drag source when a drag has completed.
 */
export declare abstract class DragWidget extends DragDropWidgetBase {
    /**
     * Construct a drag widget.
     */
    constructor(options?: DragWidget.IOptions);
    /**
     * No-op on DragWidget, as it does not support dropping
     */
    protected processDrop(dropTarget: HTMLElement, event: Drag.Event): void;
    /**
     * Simply returns null for DragWidget, as it does not support dropping
     */
    protected findDropTarget(input: HTMLElement, mimeData: MimeData): HTMLElement | null;
}
/**
 * A widget which allows the user to rearrange widgets in the widget by
 * drag and drop. An internal drag and drop of a widget will cause it
 * to be inserted (by `insertWidget`) in the index of the widget it was
 * dropped on.
 *
 * Any descendant element with the drag handle class `'jfs-mod-dragHandle'`
 * will serve as a handle that can be used for dragging. If DragWidgets are
 * nested, handles will only belong to the closest parent DragWidget. For
 * convenience, the functions `makeHandle`, `unmakeHandle` and
 * `createDefaultHandle` can be used to indicate which elements should be
 * made handles. `createDefaultHandle` will create a new element as a handle
 * with a default styling class applied. Optionally, `childrenAreDragHandles`
 * can be set to indicate that all direct children are themselve drag handles.
 *
 * The functionallity of the class can be extended by overriding the following
 * functions:
 *  - addMimeData: Override to add other drag data to the mime bundle.
 *    This is often a necessary step for allowing dragging to external
 *    drop targets.
 *  - processDrop: Override if you need to handle other mime data than the
 *    default. For allowing drops from external sources, the field
 *    `acceptDropsFromExternalSource` should be set as well.
 *  - findDragTarget(): Override if anything other than the direct children
 *    of the widget's node are to be drag targets.
 *  - findDropTarget(): Override if anything other than the direct children
 *    of the widget's node are to be the drop targets.
 *  - getIndexOfChildNode(): Override to change the key used to represent
 *    the drag and drop target (default is index of child widget).
 *  - move(): Override to change how a move is handled.
 *  - getDragImage: Override to change the drag image (the default is a
 *    copy of the drag target).
 *  - onDragComplete(): Callback on drag source when a drag has completed.
 *
 * To drag and drop other things than all direct children, the following functions
 * should be overriden: `findDragTarget`, `findDropTarget` and possibly
 * `getIndexOfChildNode` and `move` to allow for custom to/from keys.
 *
 * For maximum control, `startDrag` and `evtDrop` can be overriden.
 */
export declare abstract class DragDropWidget extends DragDropWidgetBase {
    protected abstract move(mimeData: MimeData, target: HTMLElement): Drag.DropAction;
    /**
     * Adds mime data represeting the drag data to the drag event's MimeData bundle.
     *
     * The default implementation adds mime data indicating the index of the direct
     * child being dragged (as indicated by findDragTarget).
     *
     * Override this method if you have data that cannot be communicated well by an
     * index, for example if the data should be able to be dropped on an external
     * target that only understands direct mime data.
     *
     * As the method simply adds mime data for a specific key, overriders can call
     * this method before/after adding their own mime data to still support default
     * dragging behavior.
     */
    protected abstract addMimeData(handle: HTMLElement, mimeData: MimeData): void;
    /**
     * Processes a drop event.
     *
     * This function is called after checking:
     *  - That the `dropTarget` is a valid drop target
     *  - The value of `event.source` if `acceptDropsFromExternalSource` is false
     *
     * The default implementation assumes calling `getIndexOfChildNode` with
     * `dropTarget` will be valid. It will call `move` with that index as `to`,
     * and the index stored in the mime data as `from`.
     *
     * Override this if you need to handle other mime data than the default.
     */
    protected processDrop(dropTarget: HTMLElement, event: Drag.Event): void;
}
/**
 * The namespace for the `DropWidget` class statics.
 */
export declare namespace DropWidget {
    /**
     * An options object for initializing a drag widget widget.
     */
    interface IOptions extends Widget.IOptions {
        /**
         * Whether the lsit should accept drops from an external source.
         * Defaults to false.
         *
         * This option only makes sense to set for subclasses that accept drops from
         * external sources.
         */
        acceptDropsFromExternalSource?: boolean;
        /**
         * Which mimetypes are acceptable for drops
         */
        acceptedDropMimeTypes?: string[];
    }
    /**
     * Validate a drop action against a SupportedActions type
     */
    function isValidAction(supported: Drag.SupportedActions, action: Drag.DropAction): boolean;
}
/**
 * The namespace for the `DragWidget` class statics.
 */
export declare namespace DragWidget {
    /**
     * An options object for initializing a drag widget widget.
     */
    interface IOptions extends Widget.IOptions {
    }
    /**
     * Mark a widget as a drag handle.
     *
     * Using this, any child-widget can be a drag handle, as long as mouse events
     * are propagated from it to the DragWidget.
     */
    function makeHandle(handle: Widget): void;
    /**
     * Unmark a widget as a drag handle
     */
    function unmakeHandle(handle: Widget): void;
    /**
     * Create a default handle widget for dragging (see styling in DragWidget.css).
     *
     * The handle will need to be styled to ensure a minimum size
     */
    function createDefaultHandle(): Widget;
}
/**
 * The namespace for the `DragDropWidget` class statics.
 */
export declare namespace DragDropWidget {
    interface IOptions extends DragWidget.IOptions, DropWidget.IOptions {
    }
}
