/******************************************************************************
 *
 * Copyright (c) 2023, the jupyter-fs authors.
 *
 * This file is part of the jupyter-fs library, distributed under the terms of
 * the Apache License 2.0.  The full license can be found in the LICENSE file.
 *
 */
import { URLExt } from "@jupyterlab/coreutils";
import { ServerConnection } from "@jupyterlab/services";
import * as React from "react";
import { splitPathstrDrive } from "./contents_utils";
function _mknode(obj, paths) {
    var _a;
    for (const path of paths) {
        obj = obj[path] = (_a = obj[path]) !== null && _a !== void 0 ? _a : {};
    }
    return obj;
}
/**
 * Trick to set uiSchema on our settings editor form elements.
 *
 * We use it to set the "template" to a "textarea" multiline input
 */
export function snippetFormRender(props) {
    const ArrayField = props.registry.fields.ArrayField;
    const uiSchema = { ...props.uiSchema };
    const templateUiSchema = _mknode(uiSchema, ["items", "template"]);
    templateUiSchema["ui:widget"] = "textarea";
    return React.createElement(ArrayField, { ...props, uiSchema: uiSchema });
}
const processUrlRegex = /^(?<protocol>.+?):\/\/(?:[^@]+@)?(?<resource>.*)$/;
const templateTokenFinders = {};
/**
 * Get all the snippet specifications in the settings
 */
export function getSettingsSnippets(settings) {
    var _a;
    const raw = ((_a = settings === null || settings === void 0 ? void 0 : settings.composite.snippets) !== null && _a !== void 0 ? _a : []);
    return raw.map(s => ({ ...s, pattern: new RegExp(s.pattern) }));
}
/**
 * Gets all the snippet specifications configured on the server
 */
export async function getServerSnippets(settings) {
    if (!settings) {
        settings = ServerConnection.makeSettings();
    }
    return ServerConnection.makeRequest(URLExt.join(settings.baseUrl, "/jupyterfs/snippets"), { method: "GET" }, settings).then(response => {
        if (!response.ok) {
            return Promise.reject(response);
        }
        return response.json();
    }).then(data => data.snippets.map(s => ({ ...s, pattern: new RegExp(s.pattern) })));
}
/**
 * Get the snippet specifications from all sources
 */
export async function getAllSnippets(settings) {
    return (await getServerSnippets()).concat(getSettingsSnippets(settings));
}
/**
 * Instantiate the template of a snippet.
 *
 * @param template The template to instantiate
 * @param resource The resource the entry belongs to
 * @param path The local path of the entry
 */
export function instantiateSnippet(template, url, type, pathstr) {
    const parsed = processUrlRegex.exec(url);
    const [drive, relativePath] = splitPathstrDrive(pathstr);
    const args = {
        ...parsed === null || parsed === void 0 ? void 0 : parsed.groups,
        url,
        type,
        path: relativePath,
        full_url: `${url.replace(/\/$/, "")}/${relativePath}`,
        full_path: `${drive}:/${relativePath}`,
        drive,
    };
    let templated = template;
    for (const key of Object.keys(args)) {
        if (!(key in templateTokenFinders)) {
            // match `key` wrapped in double curly-braces (and optionally whitespace padding within the braces)
            templateTokenFinders[key] = new RegExp(`\\{\\{\\s*${key}\\s*\\}\\}`, "g");
        }
        templated = templated.replace(templateTokenFinders[key], args[key]);
    }
    return templated;
}
//# sourceMappingURL=snippets.js.map