import { Dialog, showDialog, showErrorMessage, Toolbar, ToolbarButton, WidgetTracker, /*Clipboard, Dialog, IWindowResolver, showDialog*/ } from "@jupyterlab/apputils";
// import { PathExt, URLExt } from "@jupyterlab/coreutils";
import { isValidFileName /*renameFile*/ } from "@jupyterlab/docmanager";
import { nullTranslator, } from "@jupyterlab/translation";
import { refreshIcon, newFolderIcon, } from "@jupyterlab/ui-components";
// import JSZip from "jszip";
import { ArrayExt } from "@lumino/algorithm";
import { PromiseDelegate } from "@lumino/coreutils";
import { PanelLayout, Widget } from "@lumino/widgets";
import { Format, Path } from "@tree-finder/base";
import { JupyterClipboard } from "./clipboard";
import { commandIDs, idFromResource } from "./commands";
import { ContentsProxy } from "./contents_proxy";
import { getContentParent, revealPath, openDirRecursive } from "./contents_utils";
import { DragDropWidget, TABLE_HEADER_MIME } from "./drag";
import { fileTreeIcon } from "./icons";
import { promptRename } from "./utils";
import { Uploader, UploadButton } from "./upload";
export class TreeFinderTracker extends WidgetTracker {
    constructor() {
        super(...arguments);
        this._finders = new Map();
    }
    async add(finder) {
        this._finders.set(finder.id, finder);
        // eslint-disable-next-line @typescript-eslint/unbound-method
        finder.disposed.connect(this._onWidgetDisposed, this);
        return super.add(finder);
    }
    remove(finder) {
        this._finders.delete(finder.id);
        // eslint-disable-next-line @typescript-eslint/unbound-method
        finder.disposed.disconnect(this._onWidgetDisposed, this);
    }
    findByDrive(drive) {
        return this._finders.get(drive);
    }
    hasByDrive(drive) {
        return this._finders.has(drive);
    }
    _onWidgetDisposed(finder) {
        this.remove(finder);
    }
}
export class TreeFinderWidget extends DragDropWidget {
    constructor({ app, columns, rootPath = "", translator, settings, }) {
        const { commands, serviceManager: { contents } } = app;
        const node = document.createElement("tree-finder-panel");
        const acceptedDropMimeTypes = [TABLE_HEADER_MIME];
        super({ node, acceptedDropMimeTypes });
        this.addClass("jp-tree-finder");
        // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
        this.contentsProxy = new ContentsProxy(contents, rootPath, this.onGetChildren.bind(this));
        this.settings = settings;
        this.translator = translator || nullTranslator;
        this._trans = this.translator.load("jupyterlab");
        this._commands = commands;
        this._expanding = new Map();
        this._columns = columns;
        this.rootPath = rootPath === "" ? rootPath : rootPath + ":";
        this._initialLoad = true;
        this._readyDelegate = new PromiseDelegate();
        void this._readyDelegate.promise.catch(reason => showErrorMessage("Failed to init browser", reason));
        void this._readyDelegate.promise.then(() => {
            // TODO: Model state of TreeFinderWidget should be updated by renamerSub process.
            //       Currently we hard-code the refresh here, but should be moved upstream!
            const contentsModel = this.model;
            contentsModel.renamerSub.subscribe(({ name, target }) => {
                void contentsModel.sort();
            });
        });
    }
    move(mimeData, target) {
        const source = mimeData.getData(TABLE_HEADER_MIME);
        const dest = target.innerText;
        void this._reorderColumns(source, dest);
        void this.nodeInit();
        return "move";
    }
    addMimeData(handle, mimeData) {
        const columnName = handle.innerText;
        mimeData.setData(TABLE_HEADER_MIME, columnName);
    }
    getDragImage(handle) {
        const target = this.findDragTarget(handle);
        let img = null;
        if (target) {
            img = target.cloneNode(true);
            img.classList.add("jp-thead-drag-image");
        }
        return img;
    }
    onGetChildren(path, done) {
        if (this._initialLoad) {
            const rootPathstr = Path.toarray(this.rootPath).join("/");
            if (path === rootPathstr) {
                done.finally(() => {
                    this._initialLoad = false;
                    this.draw();
                });
            }
        }
        this._expanding.set(path, (this._expanding.get(path) || 0) + 1);
        // only redraw if bumped up from 0
        if (this._expanding.get(path) === 1) {
            this.draw();
        }
        done.finally(() => {
            this._expanding.set(path, (this._expanding.get(path) || 1) - 1);
            // only redraw if bumped down to 0
            if (this._expanding.get(path) === 0) {
                this.draw();
            }
        });
    }
    /**
     * Reorders the columns according to given inputs and saves to user settings
     * If `source` is dragged from left to right, it will be inserted to the right side of `dest`
     * Else if `source` is dragged from right to left, it will be inserted to the left side of `dest`
     */
    async _reorderColumns(source, dest) {
        var _a;
        const sIndex = this._columns.indexOf(source);
        const dIndex = this._columns.indexOf(dest);
        if (sIndex < dIndex) {
            this._columns.splice(dIndex + 1, 0, source);
            this._columns.splice(sIndex, 1);
        }
        else if (sIndex > dIndex) {
            this._columns.splice(sIndex, 1);
            this._columns.splice(dIndex, 0, source);
        }
        await ((_a = this.settings) === null || _a === void 0 ? void 0 : _a.set("display_columns", this._columns));
    }
    draw() {
        var _a;
        (_a = this.model) === null || _a === void 0 ? void 0 : _a.requestDraw();
    }
    refresh() {
        var _a;
        (_a = this.model) === null || _a === void 0 ? void 0 : _a.refreshSub.next([]);
    }
    async nodeInit() {
        var _a;
        // The contents of root passed to node.init is not (currently) considered, so do not ask for it.
        let root = null;
        try {
            root = await this.contentsProxy.get(this.rootPath, { content: false });
        }
        catch (error) {
            return;
        }
        this._currentFolder = (_a = this.model) === null || _a === void 0 ? void 0 : _a.root.pathstr;
        await this.node.init({
            root,
            gridOptions: {
                columnFormatters: {
                    last_modified: (x => Format.timeSince(x)),
                    size: (x => x && Format.bytesToHumanReadable(x)),
                    writable: (x => x ? "✓" : "╳"),
                },
                doWindowResize: true,
                showFilter: true,
            },
            modelOptions: {
                columnNames: this.columns,
            },
        });
        const grid = this.node.querySelector("tree-finder-grid");
        grid === null || grid === void 0 ? void 0 : grid.addStyleListener(() => {
            var _a, _b, _c;
            // Set root-level load indicator
            grid.classList.toggle("jfs-mod-loading", this._initialLoad);
            // Fix corner cleanup (workaround for underlying bug where we end up with two resize handles)
            const resizeSpans = grid.querySelectorAll(`thead tr > th:first-child > span.rt-column-resize`);
            const nHeaderRows = grid.querySelectorAll("thead tr").length;
            if (resizeSpans.length > nHeaderRows) {
                // something went wrong, and we ended up with double resize handles. Clear the classes from the first one:
                for (const span of grid.querySelectorAll(`thead tr > th:first-child > span.rt-column-resize:first-child`)) {
                    span.removeAttribute("class");
                }
            }
            // Fix focus and tabbing
            let lastSelectIdx = ((_a = this.model) === null || _a === void 0 ? void 0 : _a.selectedLast) ? (_b = this.model) === null || _b === void 0 ? void 0 : _b.contents.indexOf(this.model.selectedLast) : -1;
            const lostFocus = document.activeElement === document.body;
            for (const rowHeader of grid.querySelectorAll("tr > th")) {
                const tableHeader = rowHeader.querySelector("span.tf-header-name");
                if (tableHeader) {
                    // If tableheader is path, do not make it draggable
                    if (tableHeader.innerText !== "path") {
                        tableHeader.classList.add(this.dragHandleClass);
                    }
                }
                const nameElement = rowHeader.querySelector("span.rt-group-name");
                // Ensure we can tab to all items
                nameElement === null || nameElement === void 0 ? void 0 : nameElement.setAttribute("tabindex", "0");
                // Ensure last selected element retains focus after redraw:
                if (lostFocus && nameElement && lastSelectIdx !== -1) {
                    const meta = grid.getMeta(rowHeader);
                    if (meta && meta.y === lastSelectIdx) {
                        nameElement.focus();
                        lastSelectIdx = -1;
                    }
                }
                // Add "loading" indicator for folders that are fetching children
                if (nameElement) {
                    const meta = grid.getMeta(rowHeader);
                    const content = (meta === null || meta === void 0 ? void 0 : meta.y) ? (_c = this.model) === null || _c === void 0 ? void 0 : _c.contents[meta.y] : undefined;
                    if (content) {
                        rowHeader.classList.toggle("jfs-mod-loading", !!nameElement && (this._expanding.get(content.pathstr) || 0) > 0);
                    }
                }
            }
        });
        if (this.uploader) {
            this.uploader.model = this.model;
        }
        else {
            this.uploader = new Uploader({
                contentsProxy: this.contentsProxy,
                model: this.model,
            });
        }
        if (this._currentFolder) {
            await openDirRecursive(this.model, this._currentFolder.split("/"));
        }
        this.model.openSub.subscribe(rows => rows.forEach(row => {
            if (!row.getChildren) {
                void this._commands.execute("docmanager:open", { path: Path.fromarray(row.path) });
            }
            else {
                const widget = TreeFinderSidebar.tracker.findByDrive(this.parent.id);
                void TreeFinderSidebar.tracker.save(widget);
            }
        }));
    }
    get columns() {
        return this._columns;
    }
    set columns(value) {
        if (ArrayExt.shallowEqual(this._columns, value)) {
            return;
        }
        this._columns = value;
        const m = this.model;
        m.options = {
            ...m.options,
            columnNames: this._columns,
        };
        m.initColumns();
        void this.nodeInit();
    }
    get ready() {
        return this._readyDelegate.promise;
    }
    get model() {
        return this.node.model;
    }
    get selection() {
        var _a;
        return (_a = this.model) === null || _a === void 0 ? void 0 : _a.selection;
    }
    get selectionPathstrs() {
        var _a;
        return (_a = this.model) === null || _a === void 0 ? void 0 : _a.selection.map(c => Path.fromarray(c.row.path));
    }
    /**
     * Handle the DOM events for the tree view.
     *
     * @param event - The DOM event sent to the widget.
     *
     * #### Notes
     * This method implements the DOM `EventListener` interface and is
     * called in response to events on the panel's DOM node. It should
     * not be called directly by user code.
     */
    handleEvent(event) {
        switch (event.type) {
            case "keydown":
                this.evtKeydown(event);
                break;
            case "dragenter":
                this.evtNativeDragOverEnter(event);
            case "dragover":
                this.evtNativeDragOverEnter(event);
                break;
            case "dragleave":
            case "dragend":
                this.evtNativeDragLeaveEnd(event);
                break;
            case "drop":
                this.evtNativeDrop(event);
                break;
            default:
                super.handleEvent(event);
                break;
        }
    }
    /**
     * A message handler invoked on an `'after-attach'` message.
     */
    onAfterAttach(msg) {
        super.onAfterAttach(msg);
        const node = this.node;
        const initPromise = this.nodeInit();
        if (this._initialLoad) {
            void initPromise.then(() => {
                this._readyDelegate.resolve();
            });
        }
        node.addEventListener("keydown", this);
        node.addEventListener("dragenter", this);
        node.addEventListener("dragover", this);
        node.addEventListener("dragleave", this);
        node.addEventListener("dragend", this);
        node.addEventListener("drop", this);
    }
    /**
     * A message handler invoked on a `'before-detach'` message.
     */
    onBeforeDetach(msg) {
        super.onBeforeDetach(msg);
        const node = this.node;
        node.removeEventListener("keydown", this);
        node.removeEventListener("dragover", this);
        node.removeEventListener("dragover", this);
        node.removeEventListener("dragleave", this);
        node.removeEventListener("dragend", this);
        node.removeEventListener("drop", this);
    }
    _findEventRowElement(event, selector) {
        let node = event.target;
        while (node.parentElement && node.parentElement !== this.node) {
            if (node.matches(selector)) {
                return node;
            }
            node = node.parentElement;
        }
    }
    selectNeighbour(diff, asRange) {
        const model = this.model;
        let last = model.selectedLast;
        // tree-finder has a bug where it doesn't update selectedLast in `selectRange`, hacky work-around for now:
        const { range, pivot } = model.selectionModel;
        // once bug is fixed, this conditional should never be true:
        if (pivot && pivot === (last === null || last === void 0 ? void 0 : last.pathstr) && range && range.length >= 1) {
            // get the part of range that is furthest away from pivot:
            const paths = model.contents.map(c => c.pathstr);
            const pivotIdx = paths.indexOf(pivot);
            const rangeStartIdx = paths.indexOf(range[0]);
            const rangeEndIdx = paths.indexOf(range[range.length - 1]);
            if (pivotIdx < rangeStartIdx) {
                last = model.contents[rangeEndIdx];
            }
            else {
                last = model.contents[rangeStartIdx];
            }
        }
        let idx = last
            ? model.contents.indexOf(last)
            : diff < 0
                ? model.contents.length - 1 // select last item
                : 0; // select first item
        if (last) {
            idx = idx + diff;
        }
        if (idx < 0 || idx >= model.contents.length) {
            return; // Do nothing if going past the edge
        }
        const next = model.contents[idx];
        if (asRange) {
            model.selectionModel.selectRange(next, model.contents);
        }
        else {
            model.selectionModel.select(next);
        }
        void TreeFinderSidebar.scrollIntoView(this, next.pathstr);
    }
    evtKeydown(event) {
        var _a, _b, _c, _d;
        // handle any keys unaffacted by renaming status above this check:
        if ((_a = this.parent) === null || _a === void 0 ? void 0 : _a.node.classList.contains("jfs-mod-renaming")) {
            return;
        }
        switch (event.key) {
            case "ArrowDown":
            case "ArrowUp":
                event.stopPropagation();
                event.preventDefault();
                this.selectNeighbour(event.key === "ArrowUp" ? -1 : 1, event.shiftKey);
                break;
            case "ArrowLeft":
                if ((_b = this.model) === null || _b === void 0 ? void 0 : _b.selectedLast) {
                    event.stopPropagation();
                    event.preventDefault();
                    const selectedLast = this.model.selectedLast;
                    // don't allow expansion or up/down nav if in select range mode:
                    if (!event.shiftKey) {
                        if (selectedLast.isExpand) {
                            void this.model.collapse(this.model.contents.indexOf(selectedLast));
                        }
                        else {
                            // navigate the selection to the next up (exluding to root)
                            void getContentParent(selectedLast, this.model.root).then(parent => {
                                var _a, _b;
                                if (parent !== ((_a = this.model) === null || _a === void 0 ? void 0 : _a.root)) {
                                    (_b = this.model) === null || _b === void 0 ? void 0 : _b.selectionModel.select(parent);
                                    return TreeFinderSidebar.scrollIntoView(this, parent.pathstr);
                                }
                            });
                        }
                    }
                }
                break;
            case "ArrowRight":
                if ((_c = this.model) === null || _c === void 0 ? void 0 : _c.selectedLast) {
                    event.stopPropagation();
                    event.preventDefault();
                    const selectedLast = this.model.selectedLast;
                    // don't allow expansion or up/down nav if in select range mode:
                    if (!event.shiftKey) {
                        if (!selectedLast.isExpand) {
                            void this.model.expand(this.model.contents.indexOf(selectedLast));
                        }
                        else if (selectedLast.hasChildren) {
                            // navigate the selection to the first child
                            void selectedLast.getChildren().then(children => {
                                var _a;
                                if (children && children.length > 0) {
                                    (_a = this.model) === null || _a === void 0 ? void 0 : _a.selectionModel.select(children[0]);
                                    return TreeFinderSidebar.scrollIntoView(this, children[0].pathstr);
                                }
                            });
                        }
                    }
                }
                break;
            case " ": // space key
                // Toggle expansion if dir
                if ((_d = this.model) === null || _d === void 0 ? void 0 : _d.selectedLast) {
                    event.stopPropagation();
                    event.preventDefault();
                    const selectedLast = this.model.selectedLast;
                    if (selectedLast.hasChildren) {
                        const selectedIdx = this.model.contents.indexOf(selectedLast);
                        if (selectedLast.isExpand) {
                            void this.model.collapse(selectedIdx);
                        }
                        else {
                            void this.model.expand(selectedIdx);
                        }
                    }
                }
                break;
        }
    }
    evtNativeDragOverEnter(event) {
        const row = this._findEventRowElement(event, "tree-finder-grid tr");
        if (row) {
            row.classList.add("jfs-mod-native-drop");
        }
        event.preventDefault();
    }
    evtNativeDragLeaveEnd(event) {
        const row = this._findEventRowElement(event, ".jfs-mod-native-drop");
        if (row) {
            row.classList.remove("jfs-mod-native-drop");
        }
    }
    /**
     * Handle the `drop` event for the widget.
     */
    evtNativeDrop(event) {
        var _a, _b, _c;
        const row = this.node.querySelector(".jfs-mod-native-drop");
        if (row) {
            row.classList.remove("jfs-mod-native-drop");
        }
        const files = (_a = event.dataTransfer) === null || _a === void 0 ? void 0 : _a.files;
        if (!files || files.length === 0) {
            return;
        }
        const length = (_b = event.dataTransfer) === null || _b === void 0 ? void 0 : _b.items.length;
        if (!length) {
            return;
        }
        for (let i = 0; i < length; i++) {
            const entry = (_c = event.dataTransfer) === null || _c === void 0 ? void 0 : _c.items[i].webkitGetAsEntry();
            if (entry === null || entry === void 0 ? void 0 : entry.isDirectory) {
                void showDialog({
                    title: this._trans.__("Error Uploading Folder"),
                    body: this._trans.__("Drag and Drop is currently not supported for folders"),
                    buttons: [Dialog.cancelButton({ label: this._trans.__("Close") })],
                });
            }
        }
        event.preventDefault();
        // Translate row element to contents row
        let target = this.model.root;
        if (row) {
            const grid = this.node.querySelector("tree-finder-grid");
            const metadata = grid.getMeta(row.querySelector("th"));
            if (metadata.y) {
                target = this.model.contents[metadata.y];
            }
        }
        for (let i = 0; i < files.length; i++) {
            void this.uploader.upload(files[i], target);
        }
    }
}
export class TreeFinderSidebar extends Widget {
    constructor({ app, columns, url, type, rootPath = "", caption = "TreeFinder", id = "jupyterlab-tree-finder", settings, preferredDir, }) {
        super();
        this.id = id;
        this.node.classList.add("jfs-mod-notRenaming");
        this.url = url;
        this.type = type;
        this.title.icon = fileTreeIcon;
        this.title.caption = caption;
        this.addClass("jp-tree-finder-sidebar");
        this.toolbar = new Toolbar();
        this.toolbar.addClass("jp-tree-finder-toolbar");
        this.preferredDir = preferredDir;
        this.treefinder = new TreeFinderWidget({ app, rootPath, columns, settings });
        this.layout = new PanelLayout();
        this.layout.addWidget(this.toolbar);
        this.layout.addWidget(this.treefinder);
    }
    restore() {
        void this.treefinder.ready.then(() => this.treefinder.refresh());
    }
    async download(path, folder) {
        if (folder) {
            // const zip = new JSZip();
            // await this.wrapFolder(zip, path); // folder packing
            // // generate and save zip, reset path
            // path = PathExt.basename(path);
            // writeZipFile(zip, path);
        }
        else {
            const url = await this.treefinder.contentsProxy.downloadUrl(path);
            const element = document.createElement("a");
            element.setAttribute("href", url);
            element.setAttribute("download", "");
            document.body.appendChild(element);
            element.click();
            document.body.removeChild(element);
        }
    }
    // async wrapFolder(zip: JSZip, path: string) {
    //   const base = this.cm.get(this.basepath + path);
    //   const next = base.then(async res => {
    //     if (res.type === "directory") {
    //       const new_folder = zip.folder(res.name);
    //       for (const c in res.content) {
    //         await this.wrapFolder(new_folder, res.content[c].path);
    //       }
    //     } else {
    //       zip.file(res.name, res.content);
    //     }
    //   });
    //   await next;
    // }
    onBeforeShow(msg) {
        this.treefinder.refresh();
        this.treefinder.draw();
    }
    onResize(msg) {
        this.treefinder.draw();
    }
}
// eslint-disable-next-line @typescript-eslint/no-namespace
(function (TreeFinderSidebar) {
    const namespace = "jupyter-fs-treefinder";
    TreeFinderSidebar.tracker = new TreeFinderTracker({ namespace });
    TreeFinderSidebar.clipboard = new JupyterClipboard(TreeFinderSidebar.tracker);
    function sidebarFromResource(resource, props) {
        return sidebar({
            ...props,
            rootPath: resource.drive,
            caption: resource.name,
            id: idFromResource(resource),
            preferredDir: resource.preferredDir,
            url: resource.url,
            type: resource.type,
        });
    }
    TreeFinderSidebar.sidebarFromResource = sidebarFromResource;
    function sidebar({ app, 
    // manager,
    // paths,
    // resolver,
    // router,
    restorer, url, type, columns, settings, preferredDir, rootPath = "", caption = "TreeFinder", id = "jupyterlab-tree-finder", side = "left", }) {
        const widget = new TreeFinderSidebar({ app, rootPath, columns, caption, id, url, type, settings, preferredDir });
        void widget.treefinder.ready.then(() => TreeFinderSidebar.tracker.add(widget));
        app.shell.add(widget, side);
        const new_file_button = new ToolbarButton({
            icon: newFolderIcon,
            onClick: () => {
                void app.commands.execute((commandIDs.create_folder));
            },
            tooltip: "New Folder",
        });
        const uploader_button = new UploadButton({ uploader: widget.treefinder.ready.then(() => widget.treefinder.uploader) });
        void widget.treefinder.ready.then(() => {
            widget.treefinder.uploader.uploadCompleted.connect((sender, args) => {
                // Do not select/scroll into view: Upload might be slow, so user might have moved on!
                // We do however want to expand the folder
                void revealPath(widget.treefinder.model, args.path).then(() => widget.treefinder.model.flatten());
            });
        });
        const refresh_button = new ToolbarButton({
            icon: refreshIcon,
            onClick: () => {
                void app.commands.execute(commandIDs.refresh);
            },
            tooltip: "Refresh",
        });
        widget.toolbar.addItem("new file", new_file_button);
        widget.toolbar.addItem("upload", uploader_button);
        widget.toolbar.addItem("refresh", refresh_button);
        if (preferredDir) {
            void widget.treefinder.ready.then(async () => {
                let path = preferredDir.split("/");
                if (preferredDir.startsWith("/")) {
                    path = path.slice(1);
                }
                path.unshift(rootPath);
                await openDirRecursive(widget.treefinder.model, path);
            });
        }
        // // remove context highlight on context menu exit
        // document.ondblclick = () => {
        //   app.commands.execute((widget.commandIDs.set_context + ":" + widget.id), { path: "" });
        // };
        // widget.node.onclick = event => {
        //   app.commands.execute((widget.commandIDs.select + ":" + widget.id), { path: "" });
        // };
        // setInterval(() => {
        //   app.commands.execute(widget.commandIDs.refresh);
        // }, 10000);
        // return a disposable containing all disposables associated
        // with this widget, ending with the widget itself
        return widget;
    }
    TreeFinderSidebar.sidebar = sidebar;
    async function doRename(widget, oldContent) {
        var _a;
        if (widget.node.classList.contains("jfs-mod-renaming")) {
            return oldContent.row;
        }
        const textNode = widget.node.querySelector("tr.tf-mod-select .rt-tree-container .rt-group-name").firstChild;
        const original = textNode.textContent.replace(/(.*)\/$/, "$1");
        const editNode = document.createElement("input");
        editNode.value = original;
        try {
            widget.node.classList.replace("jfs-mod-notRenaming", "jfs-mod-renaming");
            const newName = await promptRename(textNode, editNode, original);
            (_a = textNode.parentElement) === null || _a === void 0 ? void 0 : _a.focus();
            if (!newName || newName === oldContent.name) {
                return oldContent.row;
            }
            if (!isValidFileName(newName)) {
                void showErrorMessage("Rename Error", Error(newName + ' is not a valid name. Names must have nonzero length, and cannot include "/", "\\", or ":"'));
                return oldContent.row;
            }
            const oldPath = oldContent.getPathAtDepth(1).join("/");
            const newPath = oldPath.slice(0, -1 * original.length) + newName;
            const suffix = textNode.textContent.endsWith("/") ? "/" : "";
            let newContent;
            try {
                newContent = await widget.treefinder.contentsProxy.rename(oldPath + suffix, newPath + suffix);
            }
            catch (error) {
                if (error !== "File not renamed") {
                    void showErrorMessage("Rename Error", error);
                }
                newContent = oldContent.row;
            }
            textNode.textContent = newName + suffix;
            return newContent;
        }
        finally {
            widget.node.classList.replace("jfs-mod-renaming", "jfs-mod-notRenaming");
        }
    }
    TreeFinderSidebar.doRename = doRename;
    /**
     * If a path entry is not in view, scroll it into view
     *
     * @param treefinder The view
     * @param pathstr The entry to show
     */
    async function scrollIntoView(treefinder, pathstr) {
        // tree-finder uses rxjs bits that don't allow you to await, so to ensure sync draw:
        const model = treefinder.model;
        await model.flatten();
        const grid = treefinder.node.querySelector("tree-finder-grid");
        // eslint-disable-next-line @typescript-eslint/await-thenable
        await grid.draw();
        // Check if new row (selection) in view (if outside virtual window, it will fail):
        if (!treefinder.node.querySelector(".tf-mod-select .rt-tree-container .rt-group-name")) {
            // We need to scroll the selection into view!
            const rowIdx = model.contents.findIndex(s => s.pathstr === pathstr);
            if (rowIdx !== -1) {
                // TODO: Should we perform a minimum scroll, or do we always want entry as close to the top of the view as possible?
                await grid.scrollToCell(0, rowIdx, 1, model.contents.length);
            }
        }
    }
    TreeFinderSidebar.scrollIntoView = scrollIntoView;
})(TreeFinderSidebar || (TreeFinderSidebar = {}));
//# sourceMappingURL=treefinder.js.map