/******************************************************************************
 *
 * Copyright (c) 2019, the jupyter-fs authors.
 *
 * This file is part of the jupyter-fs library, distributed under the terms of
 * the Apache License 2.0.  The full license can be found in the LICENSE file.
 *
 */
import { ToolbarButton } from "@jupyterlab/apputils";
import { IChangedArgs } from "@jupyterlab/coreutils";
import { ITranslator } from "@jupyterlab/translation";
import { IDisposable } from "@lumino/disposable";
import { ISignal } from "@lumino/signaling";
import { Content, ContentsModel } from "@tree-finder/base";
import type { ContentsProxy } from "./contents_proxy";
/**
 * The size (in bytes) of the biggest chunk we should upload at once.
 */
export declare const CHUNK_SIZE: number;
/**
 * An upload completed to `path`.
 */
export interface IUploadedData {
    path: string[];
}
/**
* An upload progress event for a file at `path`.
*/
export interface IUploadProgress {
    path: string;
    /**
    * % uploaded [0, 1)
    */
    progress: number;
}
export declare class UploadButton extends ToolbarButton {
    /**
     * Construct a new file browser buttons widget.
     */
    constructor(options: UploadButton.IOptions);
    /**
     * The 'change' handler for the input field.
     */
    private _onInputChanged;
    /**
     * The 'click' handler for the input field.
     */
    private _onInputClicked;
    protected translator: ITranslator;
    private _trans;
    private _input;
    private _uploader;
}
export declare namespace UploadButton {
    interface IOptions {
        /**
         * The language translator.
         */
        translator?: ITranslator;
        /**
         * An optional label.
         */
        label?: string;
        /**
         * Pormise to uploader instance to handle upload logic
         */
        uploader: Promise<Uploader>;
    }
}
/**
 * A widget which provides an upload button.
 */
export declare class Uploader implements IDisposable {
    /**
     * Construct a new file browser buttons widget.
     */
    constructor(options: Uploader.IOptions);
    /**
     * A signal emitted when an upload progresses.
     */
    get uploadChanged(): ISignal<this, IChangedArgs<IUploadProgress | null>>;
    /**
     * A signal emitted when an upload completes.
     */
    get uploadCompleted(): ISignal<this, IUploadedData>;
    /**
     * Is this instance disposed?
     */
    get isDisposed(): boolean;
    /**
     * Dispose of the resources held by the model.
     */
    dispose(): void;
    /**
     * Upload a `File` object.
     *
     * @param file - The `File` object to upload.
     *
     * @returns A promise containing the new file contents model.
     */
    upload(file: File, target?: Content<ContentsProxy.IJupyterContentRow>): Promise<ContentsProxy.IJupyterContentRow | null>;
    /**
     * Perform the actual upload.
     */
    private _upload;
    private _uploadCheckDisposed;
    model: ContentsModel<ContentsProxy.IJupyterContentRow>;
    private _contentsProxy;
    private _uploads;
    private _uploadChanged;
    private _uploadCompleted;
    private _disposed;
}
/**
 * The namespace for Uploader class statics.
 */
export declare namespace Uploader {
    /**
     * The options used to create an uploader.
     */
    interface IOptions {
        /**
         * Contents model
         */
        model: ContentsModel<ContentsProxy.IJupyterContentRow>;
        /**
         * Contents manager proxy
         */
        contentsProxy: ContentsProxy;
    }
}
